@@ -1,8 +1,7 @@
 #===============================================================================
 #         FILE:  Games::Go::AGA::TDListDB
 #     ABSTRACT:  an SQL object for holding AGA TDList data
-#       AUTHOR:  Reid Augustin (REID), <reid@lucidport.com>
-#      COMPANY:  LucidPort Technology, Inc.
+#       AUTHOR:  Reid Augustin (REID), <reid@hellosix.com>
 #      CREATED:  12/02/2010 08:51:22 AM PST
 #===============================================================================
 
@@ -11,23 +10,9 @@ use strict;
 use warnings;
 
 package Games::Go::AGA::TDListDB;
-use parent 'Exporter';
-our @EXPORT_OK = qw(
-    LAST_NAME
-    FIRST_NAME
-    ID
-    MEMBERSHIP
-    RANK
-    DATE
-    CLUB
-    STATE
-    EXTRA
-);
-our %EXPORT_TAGS = (all => \@EXPORT_OK);
 
 use open qw( :utf8 :std );  # UTF8 for all files and STDIO
 use IO::Handle;     # for autoflush
-use Carp;
 use DBI;
 use Readonly;
 use Try::Tiny;
@@ -36,18 +21,47 @@ use LWP::UserAgent;
 use LWP::Protocol::https;
 use Games::Go::AGA::Parse::TDList;
 use Games::Go::AGA::Parse::Util qw( is_Rating );
+use Readonly;
 
-our $VERSION = '0.032'; # VERSION
+our $VERSION = '0.048'; # VERSION
 
-sub LAST_NAME  { 0 };
-sub FIRST_NAME { 1 };
-sub ID         { 2 };
-sub MEMBERSHIP { 3 };
-sub RANK       { 4 };
-sub DATE       { 5 };
-sub CLUB       { 6 };
-sub STATE      { 7 };
-sub EXTRA      { 8 };
+Readonly my $BUF_MAX  => 4096;
+
+# list the names of the default database field index subroutines, in order
+my @_column_names = (qw(
+    last_name
+    first_name
+    id
+    membership
+    rank
+    date
+    club
+    state
+    extra
+));
+my %_column_idx = (
+    last_name  => 0,
+    first_name => 1,
+    id         => 2,
+    membership => 3,
+    rank       => 4,
+    date       => 5,
+    club       => 6,
+    state      => 7,
+    extra      => 8,
+);
+
+# list of default column names
+sub column_idx {
+    my ($self, $which) = @_;
+
+    if (@_ >= 2) {
+        return $_column_idx{lc $which};
+    }
+    return wantarray
+        ?  @_column_names
+        : \@_column_names;
+}
 
 Readonly my @columns => (
     {last_name  => 'VARCHAR(256)',             },
@@ -198,22 +212,50 @@ sub new {
     return $self;
 }
 
+my $usage = qq(
+
+TDListDB [ -tdlist_file filename ] [ -sqlite_file filename ]
+         [ -url url | AGA ] [ -verbose ] [ -help ]
+
+Options may be abbreviated to their first letter.
+
+By default, TDListDB.pm updates from a file in the current
+directory named TDList.txt.  Specify -tdlist_file to update
+from a different file, or specify -url to update from a
+website.  -url AGA updates from the usual AGA website at
+    https://www.usgo.org/ratings/TDListN.txt
+
+);
+
 sub run {
     my ($class) = @_;
 
-    my $tdlist = $class->new();
+    require Getopt::Long;
+    Getopt::Long->import(qw( :config pass_through ));
+
+    my $verbose;
+    my $url;
+    exit 0 if (not GetOptions(
+        'tdlist_file=s', => \$pre_defaults{raw_filename},   # update from file
+        'sqlite_file=s', => \$pre_defaults{dbdname},        # sqlite file
+        'url=s',         => \$url,                          # URL to update from
+        'verbose',       => \$verbose,
+        'help'           => sub { print $usage; exit 0; },
+    ));
+
+    my $tdlist = $class->new( verbose => $verbose );
     my $filename = $tdlist->raw_filename;
     my $dbfile = $tdlist->dbdname;
     STDOUT->autoflush(1);
 
-    if ($ARGV[0]) {
-        if ($ARGV[0] eq 'AGA') {
-            my $url = $tdlist->url;
-            print "Updating $dbfile from AGA ($url)\n";
-            $tdlist->update_from_AGA();
-            exit;
+    if ($url) {
+        if (uc $url ne 'AGA') {
+            $tdlist->url($url);
         }
-        $filename = $ARGV[0] if (-f $ARGV[0]);
+        $url = $tdlist->url;
+        print "Updating $dbfile from AGA ($url)\n";
+        $tdlist->update_from_AGA();
+        exit;
     }
     print "Updating $dbfile from file ($filename)\n";
     $tdlist->update_from_file($filename);
@@ -305,7 +347,8 @@ sub extra_columns_callback {
 
     if (@_ > 1) {
         if (ref $new[0] ne 'CODE') {
-            croak("Must set a code-ref in extra_columns_callback\n");
+            $self->my_print("Must set a code-ref in extra_columns_callback\n");
+            die;
         }
         $self->{extra_columns_callback} = $new[0];
     }
@@ -343,7 +386,8 @@ sub db {
                 )
             }
             else {
-                croak("No dbdname for SQLite file\n");
+                $self->my_print("No dbdname for SQLite file\n");
+                die;
             }
         }
         $self->{db} = $new;
@@ -357,13 +401,19 @@ sub db {
 sub sth {
     my ($self, $name, $new) = @_;
 
-    croak("Statement handle name is required\n") if (not $name);
+    if (not $name) {
+        $self->my_print("Statement handle name is required\n");
+        die;
+    }
     if (@_ > 2) {
         $self->{sth}{$name} = $new;
     }
 
     my $sth = $self->{sth}{$name};
-    croak("No statement handle called '$name'\n") if (not $sth);
+    if (not $sth) {
+        $self->my_print("No statement handle called '$name'\n");
+        die;
+    }
 
     return $sth;
 }
@@ -421,7 +471,7 @@ sub select_id {
     # ID is primary index, so can only be one - fetch into first array
     # element:
     my ($player) = $self->sth('select_id')->fetchall_arrayref;
-    $player->[RANK] += 0 if (is_Rating($player->[RANK]));   # numify ratings
+    $player->[$self->column_idx('rank')] += 0 if (is_Rating($player->[$self->column_idx('rank')]));   # numify ratings
     return wantarray
         ? @{$player->[0]}
         : $player->[0];
@@ -430,7 +480,7 @@ sub select_id {
 sub insert_player {
     my ($self, @new) = @_;
 
-    $new[ID] = $self->next_tmp_id(1) if (not $new[ID]);
+    $new[$self->column_idx('id')] = $self->next_tmp_id(1) if (not $new[$self->column_idx('id')]);
     $self->sth('insert_player')->execute(@new);
     return wantarray
         ? @new
@@ -483,14 +533,11 @@ sub update_from_AGA {
     }
 
     my $fname = $self->raw_filename;
-    print "Starting $fname fetch at ", scalar localtime, "\n" if ($self->verbose);
-    $self->{ua}->mirror($self->url, $fname);
-    print "... fetch done at ", scalar localtime, "\n" if ($self->verbose);
-    my $fh;
-    open($fh, '<', $fname)
-        or croak("Error opening $fname for reading: $!\n");
-    $self->update_from_file($fh);
-    close $fh;
+    my $url = $self->url;
+    $self->my_print("Starting $fname fetch from $url at ", scalar localtime, "\n") if ($self->verbose);
+    $self->{ua}->mirror($url, $fname);
+    $self->my_print("... fetch done at ", scalar localtime, "\n") if ($self->verbose);
+    $self->update_from_file($fname);
 
     exit if (defined $pid); # exit if this is a spawned child ($pid == 0)
 }
@@ -502,49 +549,57 @@ sub update_from_file {
         my $fname = $fh;
         $fh = undef;
         if (not open($fh, '<', $fname)) {
-            croak("Error opening $fname for reading: $!\n");
+            $self->my_print("Error opening $fname for reading: $!\n");
+            die;
         }
     }
+    $self->fh($fh);
 
     my $parser = Games::Go::AGA::Parse::TDList->new();
     my $verbose = $self->verbose;
-    print "Starting database update at ", scalar localtime, "\n" if ($verbose);
+    $self->my_print("Starting database update at ", scalar localtime, "\n") if ($verbose);
     $self->db->do('BEGIN');
-    my @errors;
+    my $error_count = 0;
     my $ii = 0;
-    while (my $line = <$fh>) {
+    my $ID = $self->column_idx('id');
+    while (1) {
+        $ii++;
+        my $line = $self->next_line;
+        last if (not defined $line);
+        next if (not $line);
+
         if ($verbose) {
-            print '.' if (++$ii % 1000 == 0);
-            print "\n" if ($ii % 40000 == 0);
+            $self->my_print('.')  if ($ii % 1000 == 0);
+            $self->my_print("\n") if ($ii % 40000 == 0);
         }
         try {   # in case a line crashes, print error but continue
-#print "parse $line" if ($verbose);
+#$self->my_print("parse $line") if ($verbose);
             $parser->parse($line);
             my $update = $parser->as_array;
-            if ($update->[LAST_NAME] or $update->[FIRST_NAME]) {
+            if ($update->[$self->column_idx('last_name')] or $update->[$self->column_idx('first_name')]) {
                 push @{$update}, $self->extra_columns_callback->($self, $update);
-                if ($update->[ID]) {
-                    if ($update->[ID] =~ m/tmp/i) {
-                        croak "TMP IDs not allowed in TDList input"
+                if ($update->[$ID]) {
+                    if ($update->[$ID] =~ m/tmp/i) {
+                        die "TMP IDs not allowed in TDList input";
                     }
                 }
                 else {
-                    $self->sth('select_by_name')->execute($update->[LAST_NAME], $update->[FIRST_NAME]);
+                    $self->sth('select_by_name')->execute($update->[$self->column_idx('last_name')], $update->[$self->column_idx('first_name')]);
                     my $players = $self->sth('select_by_name')->fetchall_arrayref;
                     for my $player (@{$players}) {
-                        if ($player->[ID] =~ m/tmp/i) {
-                            $update->[ID] = $player->[ID];  # already in DB (hope it's the same guy!)
+                        if ($player->[$ID] =~ m/tmp/i) {
+                            $update->[$ID] = $player->[$ID];  # already in DB (hope it's the same guy!)
                         }
                     }
-                    if (not $update->[ID]) {
-                        $update->[ID] = $self->next_tmp_id(1);
+                    if (not $update->[$ID]) {
+                        $update->[$ID] = $self->next_tmp_id(1);
                     }
                 }
-                if ($self->select_id($update->[ID])) {
+                if ($self->select_id($update->[$ID])) {
                     # ID is already in database, do an update
                     $self->sth('update_id')->execute(
                         @{$update},     # new values for all columns
-                        $update->[ID],  # player ID (for WHERE condition)
+                        $update->[$ID],  # player ID (for WHERE condition)
                     );
                 }
                 else {
@@ -552,22 +607,84 @@ sub update_from_file {
                     $self->insert_player(@{$update});
                 }
             }
+            else {
+                die "no name parsed from $line";
+            }
         }
         catch {
-            push @errors, $_;
-            print $_ if ($verbose);
+            $error_count++;
+            my $error = $_;
+            $self->my_print("Error at line $ii: $error");
         };
-        if (@errors >= $self->max_update_errors) {
-            push @errors, "Too many errors - aborting\n";
+        if ($error_count >= $self->max_update_errors) {
+            $self->my_print("$error_count errors - aborting\n");
             last;
         }
     }
     $self->db->do('COMMIT');  # make sure we do this!
     $self->update_time(time);
-    if (@errors > 1) {
-        unshift @errors, scalar @errors . " errors during update:\n";
+}
+
+# file might not have lines.  enforce lines here
+sub next_line {
+    my ($self) = @_;
+
+    my $offset = $self->{buf_offset};
+    if ($self->{buf_end} - $offset <= 160) {
+        $self->get_fh_chunk;
+        $offset = $self->{buf_offset};
     }
-    croak(join "\n", @errors) if(@errors);
+    return if ($offset >= $self->{buf_end});
+    my $eol_idx;
+    if ($self->{has_lines}) {
+        $eol_idx = index($self->{buf}, "\n", $offset);
+        if ($eol_idx < 0) {
+            die "no EOL";       # shouldn't happen
+        }
+    }
+    else {
+        # assume 80 characters per line
+        $eol_idx = $offset + 80;
+        # but not past the end of the buffer
+        $eol_idx = $self->{buf_end} if ($eol_idx > $self->{buf_end});
+        $eol_idx--;
+    }
+    my $len = $eol_idx - $offset;
+    my $line = substr $self->{buf}, $offset, $len;
+    $self->{buf_offset} += $len + 1;
+    return $line;
+}
+
+sub fh {
+    my ($self, $new) = @_;
+
+    if (@_ > 1) {
+        $self->{fh} = $new;
+        delete $self->{has_lines};
+        if ($new and ref $new) {
+            $self->{buf} = '';
+            $self->{buf_offset} = $self->{buf_end} = 0;
+            $self->get_fh_chunk;
+            $self->{has_lines} = (index($self->{buf}, "\n") >= 0);
+        }
+    }
+    return $self->{fh};
+}
+
+sub get_fh_chunk {
+    my ($self) = @_;
+
+    # how much is still left?
+    my $left = $self->{buf_end} - $self->{buf_offset};
+    # shift unused part of buf down to the beginning
+    substr($self->{buf}, 0, $self->{buf_offset}, '');
+    # read in a new chunk
+    my $new = read $self->fh, $self->{buf}, $BUF_MAX - $left, $left;
+    if (not defined $new) {
+        die "Read error: $!";
+    }
+    $self->{buf_end} = $left + $new;
+    $self->{buf_offset} = 0;
 }
 
 # sql columns (without column types)
@@ -627,6 +744,19 @@ sub sql_insert_qs {
     );
 }
 
+sub my_print {
+    my $self = shift;
+
+    $self->print_cb->(@_);
+}
+
+sub print_cb {
+    my ($self, $new) = @_;
+
+    $self->{print_cb} = $new if (@_ > 1);
+    return $self->{print_cb} || sub { print @_ };
+}
+
 1;
 
 __END__
@@ -641,7 +771,7 @@ Games::Go::AGA::TDListDB - an SQL object for holding AGA TDList data
 
 =head1 VERSION
 
-version 0.032
+version 0.048
 
 =head1 SYNOPSIS
 
@@ -689,9 +819,9 @@ default filename is 'tdlistdb.sqlite'.
 
 =item max_update_errors => integer
 
-An B<update_from_file> or B<update_from_AGA> collects errors until this
-number is reached, at which point the update throws an exception, reporting
-all errors.  The default value is 10.
+An B<update_from_file> or B<update_from_AGA> counts errors until this
+number is reached, at which point the update gives up and throws an
+exception.  The default value is 10.
 
 =item url => 'http://where.to.find.tdlist'
 
@@ -714,12 +844,6 @@ forked process.  When false, B<update_from_AGA> blocks until complete
 NOTE: if B<background> is true, you should arrange to call the B<reap>
 method periodically, or make other arrangements to remove zombies.
 
-=item reap
-
-Reaps zombies created when B<update_from_AGA> is called with B<background>
-true (or any other zombies, for that matter).  This is a non-blocking call,
-so it is safe to call periodically.
-
 =item table_name => 'DB_table_name'
 
 The name of the database table.  An additional table (retrievable with
@@ -766,41 +890,58 @@ ref.
     extra_columns_callback => sub {
         my ($self, $columns) = @_;
         # add extra column indicating Dan or kyu
-        return '' if not $columns->[RANK];
-        return 'Dan' if Rank_to_Rating( $columns->[RANK] ) > 0;
+        return '' if not $columns->[$self->column_idx('rank')];
+        return 'Dan' if Rank_to_Rating( $columns->[$self->column_idx('rank')] ) > 0;
         return 'Kyu';
     }
 
 This function should always return exactly the number of extra columns defined in
 the extra_columns option - the returned value may be the empty string ('').
 
+=item $tdlistdb->print_cb( [ \&callback ]
+
+Set/get the print callback.  Whenever something is to be printed, the
+B<callback> function is called with the print arguments.  B<callback>
+defaults to the standard perl print function, and new B<callback> functions
+should be written to take arguments the same way print does.
+
 =back
 
-=item LAST_NAME
+=item $tdlistdb->my_print( @args )
 
-=item FIRST_NAME
+Calls the B<my_print> B<callback> with B<@args>.
 
-=item ID
+=item $tdlistdb->column_idx( [ 'name' ] )
 
-=item MEMBERSHIP
+When 'name' is defined, it is lower-cased, and the column index for 'name' (or undef
+if 'name' isn't one of the default column names) is returned.
 
-=item RANK
+When 'name' is not provided, returns an array (or ref to array in scalar context) of
+the default column names, in order.
 
-=item DATE
+These are the default columns by name:
 
-=item CLUB
+=over
 
-=item STATE
+=item last_name
 
-=item EXTRA
+=item first_name
 
-These are the columns by name.  They are exported if requested with:
+=item id
 
-  use Games::Go::AGA::TDListDB qw(LAST_NAME FIRST_NAME ...);
+=item membership
 
-or
+=item rank
 
-  use Games::Go::AGA::TDListDB qw(:all);
+=item date
+
+=item club
+
+=item state
+
+=item extra
+
+=back
 
 =item $tdlistdb->update_from_AGA( [ $force ] )
 
@@ -874,7 +1015,7 @@ be unique.
 Add a new player to the database.  @new_column_values are values for all
 the columns, both built-in and B<extra_columns>.
 
-If the ID column is non-true, this player is assigned an ID from
+If the id column is non-true, this player is assigned an ID from
 B<next_tmp_id>.
 
 Returns the @new_column_values array (or a reference to it in scalar
@@ -909,6 +1050,12 @@ Get the numeric part of the next TMP ID (but use B<next_tmp_id()> instead).
 
 Set the numeric part of the next TMP ID (but use B<next_tmp_id('use')> instead).
 
+=item reap
+
+Reaps zombies created when B<update_from_AGA> is called with B<background>
+true (or any other zombies, for that matter).  This is a non-blocking call,
+so it is safe to call periodically.
+
 =back
 
 =back
